import pandas as pd
import pytest

from wbportfolio.pms.analytics.portfolio import Portfolio


def test_get_next_weights():
    w0 = 0.3
    w1 = 0.5
    w2 = 0.2
    r0 = 0.1
    r1 = 0.05
    r2 = -0.23
    weights = [w0, w1, w2]
    returns = [r0, r1, r2]
    portfolio = Portfolio(X=pd.DataFrame([returns]), weights=pd.Series(weights))
    next_weights = portfolio.get_next_weights()

    assert next_weights[0] == pytest.approx(w0 * (r0 + 1) / (w0 * (r0 + 1) + w1 * (r1 + 1) + w2 * (r2 + 1)), abs=10e-8)
    assert next_weights[1] == pytest.approx(w1 * (r1 + 1) / (w0 * (r0 + 1) + w1 * (r1 + 1) + w2 * (r2 + 1)), abs=10e-8)
    assert next_weights[2] == pytest.approx(w2 * (r2 + 1) / (w0 * (r0 + 1) + w1 * (r1 + 1) + w2 * (r2 + 1)), abs=10e-8)


def test_get_next_weights_solve_quantization_error():
    w0 = 0.33333334
    w1 = 0.33333333
    w2 = 0.33333333
    weights = [w0, w1, w2]
    returns = [1.0, 1.0, 1.0]  # no returns
    portfolio = Portfolio(X=pd.DataFrame([returns]), weights=pd.Series(weights))
    next_weights = portfolio.get_next_weights(round_precision=8)  # no rounding as number are all 8 decimals
    assert sum(next_weights.values()) == 1.0
    next_weights = portfolio.get_next_weights(
        round_precision=7
    )  # we expect the weight to be rounded to 6 decimals, which would lead to a total sum of 0.999999

    assert next_weights[0] == 0.3333334
    assert next_weights[1] == 0.3333333
    assert next_weights[2] == 0.3333333


def test_get_estimate_net_value():
    w0 = 0.3
    w1 = 0.5
    w2 = 0.2
    r0 = 0.1
    r1 = 0.05
    r2 = -0.23
    weights = [w0, w1, w2]
    returns = [r0, r1, r2]
    portfolio = Portfolio(X=pd.DataFrame([returns]), weights=pd.Series(weights))
    current_price = 100
    net_asset_value = portfolio.get_estimate_net_value(current_price)
    return net_asset_value == current_price * (1.0 + w0 * r0 + w1 * r1 + w2 * r2)



from contextlib import suppress
from datetime import date
from decimal import Decimal

from django.contrib.messages import error, warning
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from pandas._libs.tslibs.offsets import BDay
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcompliance.viewsets.risk_management.mixins import RiskCheckViewSetMixin
from wbcore import serializers as wb_serializers
from wbcore import viewsets
from wbcore.contrib.permission.permissions import InternalUserPermissionMixin
from wbcore.metadata.configs.display.instance_display import (
    Display,
    create_simple_display,
)
from wbcore.utils.views import CloneMixin

from wbportfolio.models import AssetPosition, Order, OrderProposal
from wbportfolio.models.orders.order_proposals import (
    execute_orders_as_task,
    push_model_change_as_task,
    replay_as_task,
)
from wbportfolio.serializers import (
    OrderProposalModelSerializer,
    OrderProposalRepresentationSerializer,
    ReadOnlyOrderProposalModelSerializer,
)

from ...filters.orders import OrderProposalFilterSet
from ...order_routing import ExecutionStatus, RoutingException
from ...permissions import IsPortfolioManager
from ..mixins import UserPortfolioRequestPermissionMixin
from .configs import (
    OrderProposalButtonConfig,
    OrderProposalDisplayConfig,
    OrderProposalEndpointConfig,
    OrderProposalPortfolioEndpointConfig,
)


class OrderProposalRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:trade"
    queryset = OrderProposal.objects.all()
    serializer_class = OrderProposalRepresentationSerializer


class OrderProposalModelViewSet(CloneMixin, RiskCheckViewSetMixin, InternalUserPermissionMixin, viewsets.ModelViewSet):
    IDENTIFIER = "wbportfolio:order"
    ordering_fields = ("trade_date",)
    ordering = ("-trade_date",)
    search_fields = ("comment",)
    filterset_fields = {"trade_date": ["exact", "gte", "lte"], "status": ["exact"]}

    queryset = OrderProposal.objects.select_related("rebalancing_model", "portfolio")
    serializer_class = OrderProposalModelSerializer
    filterset_class = OrderProposalFilterSet
    display_config_class = OrderProposalDisplayConfig
    button_config_class = OrderProposalButtonConfig
    endpoint_config_class = OrderProposalEndpointConfig

    def get_serializer_class(self):
        if self.new_mode or (
            "pk" in self.kwargs and (obj := self.get_object()) and obj.status == OrderProposal.Status.DRAFT
        ):
            return OrderProposalModelSerializer
        return ReadOnlyOrderProposalModelSerializer

    # 2 methods to parametrize the clone button functionality
    def get_clone_button_serializer_class(self, instance):
        class CloneSerializer(wb_serializers.Serializer):
            clone_date = wb_serializers.DateField(
                default=(instance.trade_date + BDay(1)).date(), label="Trade Date"
            )  # we need to change the field name from the trade proposa fields, otherwise fontend conflicts
            clone_comment = wb_serializers.TextField(label="Comment")

        return CloneSerializer

    def get_clone_button_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["clone_comment"],
                ["clone_date"],
            ]
        )

    def add_messages(self, request, instance: OrderProposal | None = None, **kwargs):
        if instance:
            if instance.status == OrderProposal.Status.PENDING and instance.has_non_successful_checks:
                warning(
                    request,
                    "This order proposal cannot be approved because there is unsuccessful pre-trade checks. Please rectify accordingly and resubmit a valid order proposal",
                )
            if (
                instance.status == OrderProposal.Status.EXECUTION
                and instance.orders.exclude(shares=0, weighting=0)
                .filter(execution_status=Order.ExecutionStatus.FAILED)
                .exists()
            ):
                warning(request, "Some orders failed confirmation. Check the list for further details.")
            if instance.execution_status in [
                ExecutionStatus.REJECTED,
                ExecutionStatus.FAILED,
                ExecutionStatus.UNKNOWN,
            ]:
                warning(
                    request,
                    f"The execution status is {ExecutionStatus[instance.execution_status].label}. Detail: {instance.execution_comment}",
                )
            elif instance.can_be_executed and instance.approver == request.user.profile:
                warning(
                    request,
                    "As the approver of these orders, you are not authorized to execute them yourself. Please assign execution to another qualified individual.",
                )

    @classmethod
    def _get_risk_checks_button_title(cls) -> str:
        return "Pre-Trade Checks"

    @action(detail=True, methods=["PATCH"], permission_classes=[IsPortfolioManager])
    def reset(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        use_desired_target_weight = request.GET.get("use_desired_target_weight") == "true"
        if order_proposal.status == OrderProposal.Status.DRAFT:
            order_proposal.orders.all().update(weighting=0)
            order_proposal.reset_orders(use_desired_target_weight=use_desired_target_weight)
            return Response({"send": True})
        return Response({"status": "Order Proposal is not Draft"}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=True, methods=["PATCH"], permission_classes=[IsPortfolioManager])
    def normalize(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        total_cash_weight = Decimal(request.data.get("total_cash_weight", Decimal("0.0")))
        if order_proposal.status == OrderProposal.Status.DRAFT:
            order_proposal.normalize_orders(total_cash_weight)
            return Response({"send": True})
        return Response({"status": "Order Proposal is not Draft"}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=True, methods=["PATCH"], permission_classes=[IsPortfolioManager])
    def replay(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        if order_proposal.portfolio.is_manageable:
            replay_as_task.delay(order_proposal.id, user_id=self.request.user.id)
            return Response({"send": True})
        return Response({"status": "Order Proposal is not Draft"}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=True, methods=["PATCH"], permission_classes=[IsPortfolioManager])
    def pushmodelchange(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        only_for_portfolio_ids = list(
            map(lambda o: int(o), filter(lambda r: r, request.data.get("only_for_portfolio_ids", "").split(",")))
        )
        approve_automatically = request.data.get("approve_automatically") == "true"
        if order_proposal.status == OrderProposal.Status.APPROVED and order_proposal.portfolio.is_model:
            push_model_change_as_task.delay(
                order_proposal.id,
                request.user.id,
                only_for_portfolio_ids=only_for_portfolio_ids,
                approve_automatically=approve_automatically,
            )
            return Response({"send": True})
        return Response(
            {"status": "Order Proposal needs to be approved and linked to be a model portfolio"},
            status=status.HTTP_400_BAD_REQUEST,
        )

    @action(detail=True, methods=["PATCH"], permission_classes=[IsPortfolioManager])
    def execute(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        if order_proposal.can_execute(request.user):
            prioritize_target_weight = request.data.get("prioritize_target_weight") == "true"
            order_proposal.execution_status = ExecutionStatus.PENDING
            order_proposal.execution_comment = "Waiting for custodian confirmation"
            order_proposal.save()
            execute_orders_as_task.delay(order_proposal.id, prioritize_target_weight=prioritize_target_weight)
            return Response({"send": True})
        return Response({"status": "Order Proposal is not Draft"}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=True, methods=["PATCH"], permission_classes=[IsPortfolioManager])
    def cancelexecution(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        if order_proposal.execution_status and order_proposal.execution_status != ExecutionStatus.CANCELLED:
            try:
                if not order_proposal.cancel_rebalancing():
                    warning(
                        request,
                        "We could not cancel the rebalancing. It is probably already executed. Please refresh status or check with an administrator.",
                    )
            except (RoutingException, ValueError) as e:
                error(request, f"Could not cancel orders proposal {order_proposal}: {str(e)}")
            return Response({"send": True})
        return Response(
            {"status": "Order Proposal is not in an execution phase, therefore, it cannot be cancelled."},
            status=status.HTTP_400_BAD_REQUEST,
        )

    @action(detail=True, methods=["PATCH"], permission_classes=[IsPortfolioManager])
    def updateexecutionstatus(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        if order_proposal.execution_status:
            try:
                if not order_proposal.custodian_router:
                    raise RoutingException(
                        "There is no custodian router for this portfolio. Please check with an administrator."
                    )
                order_proposal.refresh_execution_status()
            except (RoutingException, ValueError) as e:
                error(request, f"Could not update rebalancing status: {str(e)}")
            return Response(
                {
                    "send": True,
                }
            )
        return Response(
            {"status": "Order Proposal is not in an execution phase, therefore, its status cannot be fetched."},
            status=status.HTTP_400_BAD_REQUEST,
        )

    @action(detail=True, methods=["PATCH"], permission_classes=[IsPortfolioManager])
    def refreshreturn(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        order_proposal.refresh_returns()
        return Response(
            {"status": "Returns were refreshed with success"},
            status=status.HTTP_200_OK,
        )

    @action(detail=True, methods=["PATCH"], permission_classes=[IsPortfolioManager])
    def refreshpretradechecks(self, request, pk=None):
        order_proposal = get_object_or_404(OrderProposal, pk=pk)
        if order_proposal.status == OrderProposal.Status.DRAFT:
            order_proposal.evaluate_pretrade_checks()
        return Response(
            {"status": "Evaluate pretrade checks"},
            status=status.HTTP_200_OK,
        )


class OrderProposalPortfolioModelViewSet(UserPortfolioRequestPermissionMixin, OrderProposalModelViewSet):
    endpoint_config_class = OrderProposalPortfolioEndpointConfig

    @cached_property
    def default_trade_date(self) -> date | None:
        with suppress(AssetPosition.DoesNotExist):
            return (self.portfolio.assets.latest("date").date + BDay(1)).date()

    def get_queryset(self):
        return OrderProposal.objects.filter(portfolio=self.kwargs["portfolio_id"])



from contextlib import suppress
from datetime import date
from decimal import Decimal
from typing import Iterator

from django.contrib import admin
from django.db import models
from django.db.models import (
    Sum,
)
from django.dispatch import receiver
from ordered_model.models import OrderedModel
from pandas import Timestamp
from pandas._libs.tslibs.offsets import BDay
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.io.mixins import ImportMixin
from wbcore.contrib.io.signals import post_import
from wbfdm.models import Instrument

from wbportfolio.import_export.handlers.orders import OrderImportHandler
from wbportfolio.models import Trade
from wbportfolio.models.asset import AssetPosition
from wbportfolio.models.transactions.transactions import TransactionMixin
from wbportfolio.models.utils import get_adjusted_shares
from wbportfolio.order_routing import ExecutionInstruction
from wbportfolio.pms.typing import Position as PositionDTO


class Order(TransactionMixin, ImportMixin, OrderedModel, models.Model):
    import_export_handler_class = OrderImportHandler

    ORDER_WEIGHTING_PRECISION = (
        8  # we need to match the asset position weighting. Skfolio advices using a even smaller number (5)
    )
    currency = None

    class Type(models.TextChoices):
        REBALANCE = "REBALANCE", "Rebalance"
        DECREASE = "DECREASE", "Decrease"
        INCREASE = "INCREASE", "Increase"
        BUY = "BUY", "Buy"
        SELL = "SELL", "Sell"
        NO_CHANGE = "NO_CHANGE", "No Change"  # default transaction subtype if weighing is 0

    class ExecutionStatus(models.TextChoices):
        PENDING = "PENDING", "Pending"
        CONFIRMED = "CONFIRMED", "Confirmed"
        EXECUTED = "EXECUTED", "Executed"
        FAILED = "FAILED", "Failed"
        IGNORED = "IGNORED", "Ignored"

    order_type = models.CharField(max_length=32, default=Type.BUY, choices=Type.choices, verbose_name="Trade Type")
    shares = models.DecimalField(
        max_digits=15,
        decimal_places=4,
        default=Decimal("0.0"),
        help_text="The number of shares that were traded.",
        verbose_name="Shares",
    )
    desired_target_weight = models.DecimalField(
        max_digits=9,
        decimal_places=ORDER_WEIGHTING_PRECISION,
        default=Decimal(0),
        help_text="Desired Target Weight (for compliance and audit)",
        verbose_name="Desired Target Weight",
    )
    weighting = models.DecimalField(
        max_digits=9,
        decimal_places=ORDER_WEIGHTING_PRECISION,
        default=Decimal(0),
        help_text="The weight to be multiplied against the target",
        verbose_name="Weight",
    )
    order_proposal = models.ForeignKey(
        to="wbportfolio.OrderProposal",
        related_name="orders",
        on_delete=models.CASCADE,
        help_text="The Order Proposal this trade is coming from",
    )
    daily_return = models.DecimalField(
        max_digits=ORDER_WEIGHTING_PRECISION * 2
        + 3,  # we don't expect any drift factor to be in the order of magnitude greater than 1000
        decimal_places=ORDER_WEIGHTING_PRECISION
        * 2,  # we need a higher precision for this factor to avoid float inprecision
        default=Decimal(0.0),
        verbose_name="Daily Return",
        help_text="The Ex-Post daily return",
    )
    quantization_error = models.DecimalField(
        max_digits=9,
        decimal_places=ORDER_WEIGHTING_PRECISION,
        default=Decimal(0),
        verbose_name="Quantization Error",
    )
    execution_status = models.CharField(
        max_length=12,
        default=ExecutionStatus.PENDING.value,
        choices=ExecutionStatus.choices,
        verbose_name="Execution Status",
    )
    execution_instruction = models.CharField(
        max_length=26,
        choices=ExecutionInstruction.choices,
        default=ExecutionInstruction.MARKET_ON_CLOSE.value,
        verbose_name="Execution Instruction",
    )
    execution_instruction_parameters = models.JSONField(
        default=dict, blank=True, verbose_name="Execution Instruction Parameters"
    )
    execution_comment = models.TextField(default="", blank=True, verbose_name="Execution Comment")

    execution_trade = models.OneToOneField(
        to="wbportfolio.Trade",
        related_name="order",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        help_text="The executed Trade",
    )
    order_with_respect_to = "order_proposal"

    class Meta(OrderedModel.Meta):
        verbose_name = "Order"
        verbose_name_plural = "Orders"
        indexes = [
            models.Index(fields=["order_proposal"]),
            models.Index(fields=["underlying_instrument", "value_date"]),
            models.Index(fields=["portfolio", "underlying_instrument", "value_date"]),
            models.Index(fields=["order_proposal", "underlying_instrument"]),
            # models.Index(fields=["date", "underlying_instrument"]),
        ]
        constraints = [
            models.UniqueConstraint(
                fields=["order_proposal", "underlying_instrument"],
                name="unique_order",
            ),
        ]
        # notification_email_template = "portfolio/email/trade_notification.html"

    @property
    def product(self):
        from wbportfolio.models.products import Product

        try:
            return Product.objects.get(id=self.underlying_instrument.id)
        except Product.DoesNotExist:
            return None

    @property
    @admin.display(description="Last Effective Date")
    def _last_effective_date(self) -> date:
        if hasattr(self, "last_effective_date"):
            return self.last_effective_date
        elif (
            assets := AssetPosition.unannotated_objects.filter(
                date__lt=self.value_date,
                portfolio=self.portfolio,
            )
        ).exists():
            return assets.latest("date").date

    @property
    @admin.display(description="Effective Weight")
    def _previous_weight(self) -> Decimal:
        if hasattr(self, "previous_weight"):
            return self.previous_weight
        return AssetPosition.unannotated_objects.filter(
            underlying_quote=self.underlying_instrument,
            date=self._last_effective_date,
            portfolio=self.portfolio,
        ).aggregate(s=Sum("weighting"))["s"] or Decimal(0)

    @property
    @admin.display(description="Effective Weight")
    def _effective_weight(self) -> Decimal:
        if hasattr(self, "effective_weight"):
            return self.effective_weight
        return self.order_proposal.get_orders().get(id=self.id).effective_weight

    @property
    @admin.display(description="Effective Shares")
    def _effective_shares(self) -> Decimal:
        return getattr(
            self,
            "effective_shares",
            self.get_effective_shares(),
        )

    @property
    @admin.display(description="Target Weight")
    def _target_weight(self) -> Decimal:
        return getattr(
            self, "target_weight", round(self._effective_weight + self.weighting, self.ORDER_WEIGHTING_PRECISION)
        )

    @property
    @admin.display(description="Target Shares")
    def _target_shares(self) -> Decimal:
        return getattr(self, "target_shares", self._effective_shares + self.shares)

    def __str__(self):
        ticker = f"{self.underlying_instrument.ticker}:" if self.underlying_instrument.ticker else ""
        return f"{ticker}{self.weighting}"

    def pre_save(self, load_returns: bool = True):
        self.portfolio = self.order_proposal.portfolio
        self.value_date = self.order_proposal.trade_date
        self.set_currency_fx_rate()

        if not self.price:
            self.set_price(load_returns=load_returns)

        if not self.portfolio.only_weighting and not self.shares:
            estimated_shares = self.order_proposal.get_estimated_shares(
                self.weighting, self.underlying_instrument, self.price
            )
            if estimated_shares:
                self.shares = estimated_shares
        if effective_shares := self.get_effective_shares():
            if self.order_type == self.Type.SELL:
                self.shares = -effective_shares
            else:
                self.shares = max(self.shares, -effective_shares)
        super().pre_save()

    def save(self, *args, **kwargs):
        if self.id:
            self.set_type()
        self.pre_save()
        if not self.underlying_instrument.is_investable_universe:
            self.underlying_instrument.is_investable_universe = True
            self.underlying_instrument.save()

        super().save(*args, **kwargs)

    @classmethod
    def get_type(cls, weighting, effective_weight, target_weight) -> Type:
        if weighting == 0:
            return Order.Type.NO_CHANGE
        elif weighting is not None:
            if weighting > 0:
                if abs(effective_weight) > 1e-8:
                    return Order.Type.INCREASE
                else:
                    return Order.Type.BUY
            elif weighting < 0:
                if abs(target_weight) > 1e-8:
                    return Order.Type.DECREASE
                else:
                    return Order.Type.SELL

    def get_effective_shares(self) -> Decimal:
        return AssetPosition.objects.filter(
            underlying_quote=self.underlying_instrument,
            date=self.order_proposal.last_effective_date,
            portfolio=self.portfolio,
        ).aggregate(s=Sum("shares"))["s"] or Decimal("0")

    def set_type(self):
        effective_weight = self._effective_weight
        self.order_type = self.get_type(self.weighting, effective_weight, effective_weight + self.weighting)

    def _get_price(self, load_returns: bool = True) -> tuple[Decimal, Decimal]:
        daily_return = last_price = Decimal("0")

        if self.underlying_instrument.is_cash or self.underlying_instrument.is_cash_equivalent:
            last_price = Decimal("1")
        else:
            with suppress(KeyError, TypeError):
                if load_returns:
                    self.order_proposal.portfolio.builder.load_returns(
                        [self.underlying_instrument.id],
                        self.order_proposal.last_effective_date,
                        self.value_date,
                        use_dl=True,
                    )
                last_price = Decimal(self.portfolio.builder.prices[self.value_date][self.underlying_instrument.id])
                daily_return = self.portfolio.builder.returns.loc[
                    Timestamp(self.value_date), self.underlying_instrument.id
                ]
        return last_price, daily_return

    def set_price(self, load_returns: bool = True):
        last_price, daily_return = self._get_price(load_returns=load_returns)
        self.daily_return = daily_return
        self.price = last_price

    def set_currency_fx_rate(self):
        self.currency_fx_rate = Decimal("1")
        if self.order_proposal.portfolio.currency != self.underlying_instrument.currency:
            with suppress(CurrencyFXRates.DoesNotExist):
                self.currency_fx_rate = self.underlying_instrument.currency.convert(
                    self.value_date, self.portfolio.currency, exact_lookup=True
                )

    def set_weighting(self, weighting: Decimal, portfolio_value: Decimal):
        self.weighting = weighting
        price_fx_portfolio = self.price * self.currency_fx_rate
        if price_fx_portfolio and portfolio_value:
            total_value = self.weighting * portfolio_value
            self.shares = total_value / price_fx_portfolio
        else:
            self.shares = Decimal("0")

    def set_shares(self, shares: Decimal, portfolio_value: Decimal):
        if portfolio_value:
            price_fx_portfolio = self.price * self.currency_fx_rate
            self.shares = shares
            total_value = shares * price_fx_portfolio
            self.weighting = total_value / portfolio_value
        else:
            self.weighting = self.shares = Decimal("0")

    def set_total_value_fx_portfolio(self, total_value_fx_portfolio: Decimal, portfolio_value: Decimal):
        price_fx_portfolio = self.price * self.currency_fx_rate
        if price_fx_portfolio and portfolio_value:
            self.shares = total_value_fx_portfolio / price_fx_portfolio
            self.weighting = total_value_fx_portfolio / portfolio_value
        else:
            self.weighting = self.shares = Decimal("0")

    def submit(self, by=None, description=None, portfolio_total_asset_value=None, **kwargs):
        warnings = []
        # if shares is defined and the underlying instrument defines a round lot size different than 1 and exchange allows its application, we round the share accordingly
        if self._target_weight:
            if self.order_proposal and not self.portfolio.only_weighting:
                shares = self.order_proposal.get_round_lot_size(self.shares, self.underlying_instrument)
                if shares != self.shares:
                    warnings.append(
                        f"{self.underlying_instrument.computed_str} has a round lot size of  {self.underlying_instrument.round_lot_size}: shares were rounded from {self.shares} to {shares}"
                    )
                shares = round(shares)  # ensure fractional shares are converted into integer
                # we need to recompute the delta weight has we changed the number of shares
                if shares != self.shares:
                    self.set_shares(shares, portfolio_total_asset_value)
            if abs(self.weighting) < self.order_proposal.min_weighting:
                warnings.append(
                    f"Weighting for order {self.underlying_instrument.computed_str} ({self.weighting}) is bellow the allowed Minimum Weighting ({self.order_proposal.min_weighting})"
                )
                self.set_weighting(Decimal("0"), portfolio_total_asset_value)
            if self.shares and abs(self.total_value_fx_portfolio) < self.order_proposal.min_order_value:
                warnings.append(
                    f"Total Value for order {self.underlying_instrument.computed_str} ({self.total_value_fx_portfolio}) is bellow the allowed Minimum Order Value ({self.order_proposal.min_order_value})"
                )
                self.set_weighting(Decimal("0"), portfolio_total_asset_value)
        if not self.price:
            warnings.append(f"No price for {self.underlying_instrument.computed_str}")
        if (
            not self.underlying_instrument.is_cash
            and not self.underlying_instrument.is_cash_equivalent
            and self._target_weight < -1e-8
        ):  # any value below -1e8 will be considered zero
            warnings.append(f"Negative target weight for {self.underlying_instrument.computed_str}")
        self.desired_target_weight = self._target_weight
        return warnings

    def to_dto(self) -> PositionDTO:
        return self.create_dto(
            self.underlying_instrument,
            self._target_weight,
            self.price,
            self.value_date,
            shares=self._target_shares,
            currency_fx_rate=self.currency_fx_rate,
            daily_return=self.daily_return,
        )

    @classmethod
    def create_dto(
        cls,
        instrument: Instrument,
        weighting: Decimal,
        price: Decimal,
        trade_date: date,
        shares: Decimal | None = None,
        **extra_param,
    ) -> PositionDTO:
        price_data = {}

        return PositionDTO(
            underlying_instrument=instrument.id,
            instrument_type=instrument.security_instrument_type.id,
            weighting=weighting,
            shares=shares,
            currency=instrument.currency.id,
            date=trade_date,
            asset_valuation_date=trade_date,
            is_cash=instrument.is_cash or instrument.is_cash_equivalent,
            price=price,
            exchange=instrument.exchange.id if instrument.exchange else None,
            country=instrument.country.id if instrument.country else None,
            **price_data,
            **extra_param,
        )

    @classmethod
    def handle_trade_import(cls, trade: Trade, bday_window: int = 2) -> None:
        """Link trade to earliest matching Order in business day window (default: 2 days back from book_date-1BD)."""
        if trade.underlying_instrument.is_cash or trade.is_customer_trade:
            return

        max_search_date = (trade.book_date - BDay(1)).date()  # trade are expected in t+1
        min_search_date = (max_search_date - BDay(bday_window)).date()

        instrument_filter = models.Q(underlying_instrument=trade.underlying_instrument)
        if trade.underlying_instrument.parent:
            instrument_filter |= models.Q(underlying_instrument__parent=trade.underlying_instrument.parent)
        base_qs = Order.objects.filter(
            portfolio=trade.portfolio,
            value_date__gte=min_search_date,
            value_date__lte=max_search_date,
        ).filter(instrument_filter)
        with suppress(Order.DoesNotExist):
            order = base_qs.earliest("value_date")
            if not order.execution_trade or order.execution_trade.book_date > trade.book_date:
                order.execution_trade = trade
                order.execution_status = Order.ExecutionStatus.EXECUTED
                order.save(update_fields=["execution_trade", "execution_status"])


@receiver(post_import, sender="wbportfolio.Trade")
def link_trade_to_order(sender, imported_objects, **kwargs):
    """Gets or create the fees for a given price and updates them if necessary"""
    for trade in imported_objects:
        Order.handle_trade_import(trade)


def adjust_orders(qs: Iterator[Order], underlying_quote: Instrument):
    objs = []
    for o in qs:
        old_price: Decimal = o.price
        o.underlying_instrument = underlying_quote
        o.set_price()
        if o.price and old_price != o.price and o.shares:
            o.shares = get_adjusted_shares(o.shares, old_price, o.price)
        objs.append(o)
    Order.objects.bulk_update(objs, ["price", "shares", "underlying_instrument"], batch_size=1000)
